# pass in -Gquit_at_end=true to make vsim call exit at the end. Useful for running standalone.
set quit_at_end 0
if {[lsearch $argv -Gquit_at_end=true] != -1} {
    set quit_at_end 1
}

# set up quartus_dir variable for subsequent commands
if [info exists ::env(QUARTUS_ROOTDIR_OVERRIDE)] {
    set quartus_dir $::env(QUARTUS_ROOTDIR_OVERRIDE)
} else {
    set quartus_dir "D:/altera/12.0sp2_263/quartus"
}

if {$argc > 0} {
    set base_dir $1
} else {
    set base_dir "d:/data/P4/PC_MotorControl/depot/ssg/embedded/industrial/MotorControl/ReferenceDesigns/DriveOnChip_SingleIPOneEach/dspba/./rtl"
    echo The current directory is:
    pwd
}
set base_dir [file normalize $base_dir]
echo Creating the project in $base_dir/DOC_FOC_fixp16_ax4

onerror {resume}
quit -sim
project close
project new $base_dir DOC_FOC_fixp16_ax4_FOC_FL
if {! [file exists $base_dir/work/_info]} {
    file delete -force $base_dir/work
    vlib work
}
vmap work $base_dir/work


proc checkDepends {dst src} {
    if {! [file exists $dst]} { return 1 }
    if {[file mtime $dst] < [file mtime $src]} { return 1 }
    return 0
}

set compile(altera) 0
if {[checkDepends $base_dir/altera/_info "$quartus_dir/eda/sim_lib/altera_primitives_components.vhd"]} {
    file delete -force $base_dir/altera
    vlib altera
    set compile(altera) 1
}
vmap altera $base_dir/altera
set compile(altera_mf) 0
if {[checkDepends $base_dir/altera_mf/_info "$quartus_dir/eda/sim_lib/altera_mf_components.vhd"] || [checkDepends $base_dir/altera_mf/_info "$quartus_dir/eda/sim_lib/altera_mf.vhd"]} {
    file delete -force $base_dir/altera_mf
    vlib altera_mf
    set compile(altera_mf) 1
}
vmap altera_mf $base_dir/altera_mf
set compile(lpm) 0
if {[checkDepends $base_dir/lpm/_info "$quartus_dir/eda/sim_lib/220pack.vhd"] || [checkDepends $base_dir/lpm/_info "$quartus_dir/eda/sim_lib/220model.vhd"]} {
    file delete -force $base_dir/lpm
    vlib lpm
    set compile(lpm) 1
}
vmap lpm $base_dir/lpm
set compile(wysiwyg) 0
if {[checkDepends $base_dir/wysiwyg/_info "$quartus_dir/eda/sim_lib/stratix_atoms.vhd"] || [checkDepends $base_dir/wysiwyg/_info "$quartus_dir/eda/sim_lib/stratix_components.vhd"]} {
    file delete -force $base_dir/wysiwyg
    vlib wysiwyg
    set compile(wysiwyg) 1
}
vmap wysiwyg $base_dir/wysiwyg
if {[info exists quartus_dir] && [file isdirectory $quartus_dir]} {
    if {$compile(altera)} {
        vcom -quiet -93 -work $base_dir/altera "$quartus_dir/eda/sim_lib/altera_primitives_components.vhd"
    }
    if {$compile(altera_mf)} {
        vcom -quiet -93 -work $base_dir/altera_mf "$quartus_dir/eda/sim_lib/altera_mf_components.vhd"
        vcom -quiet -93 -work $base_dir/altera_mf "$quartus_dir/eda/sim_lib/altera_mf.vhd"
    }
    if {$compile(lpm)} {
        vcom -quiet -93 -work $base_dir/lpm "$quartus_dir/eda/sim_lib/220pack.vhd"
        vcom -quiet -93 -work $base_dir/lpm "$quartus_dir/eda/sim_lib/220model.vhd"
    }
    if {$compile(wysiwyg)} {
        vcom -quiet -93 -work $base_dir/wysiwyg "$quartus_dir/eda/sim_lib/stratix_atoms.vhd"
        vcom -quiet -93 -work $base_dir/wysiwyg "$quartus_dir/eda/sim_lib/stratix_components.vhd"
    }
} else {
    echo QUARTUS not found. Not compiling libraries.
}


set vcomfailed 0
onerror {
    set vcomfailed 1
    resume
}

vcom -quiet $base_dir/DOC_FOC_fixp16_ax4/FOC/DOC_FOC_fixp16_ax4_safe_path_msim.vhd
vcom -quiet $base_dir/DOC_FOC_fixp16_ax4/FOC/FL/DOC_FOC_fixp16_ax4_FOC_FL.vhd
vcom -quiet -O0 $base_dir/DOC_FOC_fixp16_ax4/FOC/FL/DOC_FOC_fixp16_ax4_FOC_FL_atb.vhd
vcom -quiet $base_dir/DOC_FOC_fixp16_ax4/FOC/FL/DOC_FOC_fixp16_ax4_FOC_FL_stm.vhd

onerror {resume}

proc report_mismatch {signal cycle} {
    puts "Mismatch in ${signal} at system clock cycle ${cycle}"
    set modelsimvalue [examine ${signal}];
    set stmvalue [examine ${signal}_stm];
    puts "\t${signal} (ModelSim):\t${modelsimvalue}"
    puts "\t${signal} (Simulink):\t${stmvalue}"
}

if {$vcomfailed == 0} {
    onbreak {
        set my_tb [string trim [tb]];
        set regOK [regexp {(.*) ([0-9]+) ([\[address]*) ([.]*)} $my_tb \ match atbfile linenum ignore_this];
        if {$regOK == 1} {
            set simtime [expr $now - 200];
            set cyclenum [expr int($simtime / 10000.000000)];
            if { [catch {exa mismatch_qv} mismatch] == 0 && $mismatch } {
                report_mismatch qv $cyclenum
            }
            if { [catch {exa mismatch_qc} mismatch] == 0 && $mismatch } {
                report_mismatch qc $cyclenum
            }
            if { [catch {exa mismatch_valid_out} mismatch] == 0 && $mismatch } {
                report_mismatch valid_out $cyclenum
            }
            if { [catch {exa mismatch_axis_out} mismatch] == 0 && $mismatch } {
                report_mismatch axis_out $cyclenum
            }
            if { [catch {exa mismatch_Valpha} mismatch] == 0 && $mismatch } {
                report_mismatch Valpha $cyclenum
            }
            if { [catch {exa mismatch_Vbeta} mismatch] == 0 && $mismatch } {
                report_mismatch Vbeta $cyclenum
            }
            if { [catch {exa mismatch_IntegralD_out} mismatch] == 0 && $mismatch } {
                report_mismatch IntegralD_out $cyclenum
            }
            if { [catch {exa mismatch_IntegralQ_out} mismatch] == 0 && $mismatch } {
                report_mismatch IntegralQ_out $cyclenum
            }
            if { [catch {exa mismatch_debug_1} mismatch] == 0 && $mismatch } {
                report_mismatch debug_1 $cyclenum
            }
            if { [catch {exa mismatch_debug_2} mismatch] == 0 && $mismatch } {
                report_mismatch debug_2 $cyclenum
            }
            if { [catch {exa mismatch_debug_3} mismatch] == 0 && $mismatch } {
                report_mismatch debug_3 $cyclenum
            }
        } else {
            puts "Signal mismatch detected at $my_tb";
        }
        if {$quit_at_end == 1} {
            quit -code 1;
        }
    }
    vsim -voptargs="+acc" -t ps DOC_FOC_fixp16_ax4_FOC_FL_atb
    do $base_dir/DOC_FOC_fixp16_ax4/FOC/FL/DOC_FOC_fixp16_ax4_FOC_FL_atb.wav.do
    run -all
} else {
    echo At least one module failed to compile, not starting simulation
}

if {$quit_at_end == 1} {
    exit
}
