package require -exact qsys 12.0
set_module_property name dspba_auto_avalon_if_top
set_module_property AUTHOR "Altera Corporation"
set_module_property DESCRIPTION "DSP Builder Avalon-MM Interface Generator"
set_module_property DISPLAY_NAME "DSP Builder Avalon-MM Interface Generator"
set_module_property DESCRIPTION "In Qsys, the DSP Builder Avalon-MM Generator component allows a
simple interface between the DSP Builder FOC design and Qsys. The component
matches the ports in the DSP Builder design.<br /><br />
The component outputs various debugging information while parsing the DSP
Builder model. The information includes detected errors and information messages
that show the names and signal widths of the various signal types.
The component generates an Avalon-MM interface and a conduit (possibly empty)
that you should connect in your system. You should normally connect the Avalon-
MM interface to the Nios II processor and export the conduit or leave unconnected if
unnecessary.<br /><br />
The component also generates some example C code that you can use in your Nios II
software project to initialize and supply sample data to the DSP Builder design. Qsys
generates the file in the Qsys generation directory with this name:<br />
<b>\[Qsys system name\]_\[component instance name\]_regs.h</b><br /><br />
It contains address location definitions for each Avalon-MM register and the
following access functions:<br />
<b>INIT_DSPBA_\[Qsys system name\]_\[component instance name\]_regs</b><br />
- which initializes the configuration registers in the DSP Builder design<br />
<b>DSPBA_\[Qsys system name\]_\[component instance name\]_regs</b><br />
- Writes input sample data to the DSP Builder model<br />
- Starts the model running<br />
- Waits for the model to finish<br />
- Returns the output data from the DSP Builder model<br /><br />


To use the access functions, perform the following steps:<br />
1. Copy the access functions from the Qsys generation directory to the Nios Eclipse
project directory<br />
2. Include them in your C file.<br />
"

add_instance clk altera_clock_bridge
add_instance regs dspba_auto_avalon_if


add_interface clk clock end

# Error: my_component: add_instance: Can't create component type DOC_FOC_fixP16_FE_FOC_FixP16
    # while executing
# "add_instance dspba DOC_FOC_fixP16_FE_FOC_FixP16"
    # (file "X:/P4_MotorControl/depot/ssg/embedded/industrial/MotorControl/Matlab/designs/DOC_DSP/composed_example_hw.tcl" line 11)


add_instance reset altera_reset_bridge

set_interface_property clk EXPORT_OF clk.in_clk
#add_instance regs my_regs_microcore
#set_instance_property_value reset synchronous_edges deassert
add_interface reset reset end
set_interface_property reset EXPORT_OF reset.in_reset
#add_interface pins conduit end
#set_interface_property pins EXPORT_OF regs.conduit_end
add_interface slave_regs avalon slave
set_interface_property slave_regs EXPORT_OF regs.avalon_slave_0
add_connection clk.out_clk reset.clk
add_connection clk.out_clk regs.clock
add_connection reset.out_reset regs.reset

#add_connection regs.conduit_end dspba.conduit_end
#add_connection regs.output phy.input

add_parameter DSPBA_MODEL_ENTITY_NAME STRING
set_parameter_property DSPBA_MODEL_ENTITY_NAME DEFAULT_VALUE "DOC_FOC_fixp16_DOC_FOC_FixP16"
#set_parameter_property DSPBA_MODEL_ENTITY_NAME DERIVED true
set_parameter_property DSPBA_MODEL_ENTITY_NAME VISIBLE true
set_parameter_property DSPBA_MODEL_ENTITY_NAME AFFECTS_ELABORATION true
set_parameter_property DSPBA_MODEL_ENTITY_NAME AFFECTS_GENERATION true
set_parameter_property DSPBA_MODEL_ENTITY_NAME DISPLAY_NAME "Entity name of DSP Builder model"
set_parameter_property DSPBA_MODEL_ENTITY_NAME DESCRIPTION "Entity name of DSP Builder generated model. This is the name of the top level entity generated by DSP Builder. This can be found by refering to the \"set_module_property NAME \[entity name\]\" entry in the \[file name\]_hw.tcl file."


add_parameter DSPBA_MODEL_NAME STRING
set_parameter_property DSPBA_MODEL_NAME DEFAULT_VALUE "DOC_FOC_FixP16"
set_parameter_property DSPBA_MODEL_NAME VISIBLE true
set_parameter_property DSPBA_MODEL_NAME AFFECTS_ELABORATION true
set_parameter_property DSPBA_MODEL_NAME AFFECTS_GENERATION true
set_parameter_property DSPBA_MODEL_NAME DISPLAY_NAME "File name of DSP Builder model"
set_parameter_property DSPBA_MODEL_NAME DESCRIPTION "File name of DSP Builder generated model. Enter \[file name\] based on name of \[file name\]_hw.tcl file without the _hw.tcl extension."

add_parameter DSPBA_MODEL_PATH STRING
set_parameter_property DSPBA_MODEL_PATH DEFAULT_VALUE "rtl/DOC_FOC_fixp16/DOC_FOC_FixP16"
set_parameter_property DSPBA_MODEL_PATH VISIBLE true
set_parameter_property DSPBA_MODEL_PATH AFFECTS_ELABORATION true
set_parameter_property DSPBA_MODEL_PATH AFFECTS_GENERATION true
set_parameter_property DSPBA_MODEL_PATH DISPLAY_NAME "Path to DSP Builder model"
set_parameter_property DSPBA_MODEL_PATH DESCRIPTION "Pathname to DSP Builder generated \[file name\]_hw.tcl file. Path is relative to qsys project directory. Note to use '/' characters for directory separators."

add_parameter FOLDING_FACTOR NATURAL
set_parameter_property FOLDING_FACTOR DEFAULT_VALUE "50"
set_parameter_property FOLDING_FACTOR ALLOWED_RANGES "1:1000"
set_parameter_property FOLDING_FACTOR VISIBLE true
set_parameter_property FOLDING_FACTOR AFFECTS_ELABORATION true
set_parameter_property FOLDING_FACTOR AFFECTS_GENERATION true
set_parameter_property FOLDING_FACTOR DISPLAY_NAME "Folding factor of DSP Builder model"
set_parameter_property FOLDING_FACTOR DESCRIPTION "Folding factor used in DSP builder generation. Specifies the folding factor that the .m setup script uses for the DSP Builder model.
The design does not operate correctly if this does not match the DSP Builder model. Set to 1 for no folding."



add_parameter QV_NAME STRING
set_parameter_property QV_NAME DEFAULT_VALUE "qv"
set_parameter_property QV_NAME VISIBLE true
set_parameter_property QV_NAME AFFECTS_ELABORATION true
set_parameter_property QV_NAME AFFECTS_GENERATION true
set_parameter_property QV_NAME DISPLAY_NAME "Qv signal name"
set_parameter_property QV_NAME DESCRIPTION "Simulink port name for qv."
set_parameter_property QV_NAME LONG_DESCRIPTION "Simulink port name for qv. Specifies the signal names in your DSP Builder model versus the signal types for
the Avalon-MM interfaces."


add_parameter DV_NAME STRING
set_parameter_property DV_NAME DEFAULT_VALUE "dv"
set_parameter_property DV_NAME VISIBLE true
set_parameter_property DV_NAME AFFECTS_ELABORATION true
set_parameter_property DV_NAME AFFECTS_GENERATION true
set_parameter_property DV_NAME DISPLAY_NAME "Dv signal name"
set_parameter_property DV_NAME DESCRIPTION "Simulink port name for dv."
set_parameter_property DV_NAME LONG_DESCRIPTION "Simulink port name for dv. Specifies the signal names in your DSP Builder model versus the signal types for
the Avalon-MM interfaces."

add_parameter VALID_IN_NAME STRING
set_parameter_property VALID_IN_NAME DEFAULT_VALUE "valid_in"
set_parameter_property VALID_IN_NAME VISIBLE true
set_parameter_property VALID_IN_NAME AFFECTS_ELABORATION true
set_parameter_property VALID_IN_NAME AFFECTS_GENERATION true
set_parameter_property VALID_IN_NAME DISPLAY_NAME "Valid in signal name"
set_parameter_property VALID_IN_NAME DESCRIPTION "Simulink port name for Valid In (duplicate dv)."
set_parameter_property VALID_IN_NAME LONG_DESCRIPTION "Simulink port name for Valid In (duplicate dv). Specifies the signal names in your DSP Builder model versus the signal types for
the Avalon-MM interfaces."

add_parameter VALID_OUT_NAME STRING
set_parameter_property VALID_OUT_NAME DEFAULT_VALUE "valid_out"
set_parameter_property VALID_OUT_NAME VISIBLE true
set_parameter_property VALID_OUT_NAME AFFECTS_ELABORATION true
set_parameter_property VALID_OUT_NAME AFFECTS_GENERATION true
set_parameter_property VALID_OUT_NAME DISPLAY_NAME "Valid out signal name"
set_parameter_property VALID_OUT_NAME DESCRIPTION "Simulink port name for Valid Out (qv)."
set_parameter_property VALID_OUT_NAME LONG_DESCRIPTION "Simulink port name for Valid Out (qv). Specifies the signal names in your DSP Builder model versus the signal types for
the Avalon-MM interfaces."

add_parameter CHANNEL_IN_NAME STRING
set_parameter_property CHANNEL_IN_NAME DEFAULT_VALUE "axis_in"
set_parameter_property CHANNEL_IN_NAME VISIBLE true
set_parameter_property CHANNEL_IN_NAME AFFECTS_ELABORATION true
set_parameter_property CHANNEL_IN_NAME AFFECTS_GENERATION true
set_parameter_property CHANNEL_IN_NAME DISPLAY_NAME "Channel in signal name"
set_parameter_property CHANNEL_IN_NAME DESCRIPTION "Simulink port name for Channel In (dc)"
set_parameter_property CHANNEL_IN_NAME LONG_DESCRIPTION "Simulink port name for Channel In (dc). Specifies the signal names in your DSP Builder model versus the signal types for
the Avalon-MM interfaces."

add_parameter CHANNEL_OUT_NAME STRING
set_parameter_property CHANNEL_OUT_NAME DEFAULT_VALUE "axis_out"
set_parameter_property CHANNEL_OUT_NAME VISIBLE true
set_parameter_property CHANNEL_OUT_NAME AFFECTS_ELABORATION true
set_parameter_property CHANNEL_OUT_NAME AFFECTS_GENERATION true
set_parameter_property CHANNEL_OUT_NAME DISPLAY_NAME "Channel out signal name"
set_parameter_property CHANNEL_OUT_NAME DESCRIPTION "Simulink port name for Channel Out (qc)"
set_parameter_property CHANNEL_OUT_NAME LONG_DESCRIPTION "Simulink port name for Channel Out (qc). Specifies the signal names in your DSP Builder model versus the signal types for
the Avalon-MM interfaces."

add_parameter IGNORE_NAMES STRING
set_parameter_property IGNORE_NAMES DEFAULT_VALUE ""
set_parameter_property IGNORE_NAMES VISIBLE true
set_parameter_property IGNORE_NAMES AFFECTS_ELABORATION true
set_parameter_property IGNORE_NAMES AFFECTS_GENERATION true
set_parameter_property IGNORE_NAMES DISPLAY_NAME "Ignore signal names"
set_parameter_property IGNORE_NAMES DESCRIPTION "List of Simulink ports which should not be automatically connected to Avalon registers (separate multiple signals with spaces). These ports are exported as a conduit."
set_parameter_property IGNORE_NAMES LONG_DESCRIPTION "Specify signals that are not automatically added to the Avalon-MM register map and
exported as a conduit. If you specify no ignored signals, Qsys creates an empty
conduit. List of Simulink ports which should not be automatically connected to Avalon registers (separate multiple signals with spaces)."

add_parameter CFG_SUFFIX STRING
set_parameter_property CFG_SUFFIX DEFAULT_VALUE "_cfg"
set_parameter_property CFG_SUFFIX VISIBLE true
set_parameter_property CFG_SUFFIX AFFECTS_ELABORATION true
set_parameter_property CFG_SUFFIX AFFECTS_GENERATION true
set_parameter_property CFG_SUFFIX DISPLAY_NAME "Config signal name suffix"
set_parameter_property CFG_SUFFIX DESCRIPTION "All Simulink ports with this suffix will be connected to Config Input Registers. These registers are duplicated per channel."
set_parameter_property CFG_SUFFIX LONG_DESCRIPTION "If the signal name in the DSP Builder model ends with this suffix, Qsys creates them
as a bank of registers (one register per channel). To access each bank write the
channel to the channel in register. Read and write accesses to the configuration
register address go to this bank. Banks quickly swap between configuration data
when switching channels. You can set up the configuration registers for each
channel and then when switching channels with channel in, Qsys connects the
related bank of registers to the DSP Builder inputs. All Simulink ports with this suffix will be connected to Config Input Registers. These registers are duplicated per channel."

add_parameter NUM_CHANNELS INTEGER
set_parameter_property NUM_CHANNELS DEFAULT_VALUE 1
set_parameter_property NUM_CHANNELS VISIBLE true
set_parameter_property NUM_CHANNELS AFFECTS_ELABORATION true
set_parameter_property NUM_CHANNELS AFFECTS_GENERATION true
set_parameter_property NUM_CHANNELS DISPLAY_NAME "Number of channels"
set_parameter_property NUM_CHANNELS DESCRIPTION "Number of channels to support. Default is 1."
set_parameter_property NUM_CHANNELS LONG_DESCRIPTION "Number of channels to support. Default is 1."

#add_parameter PWD STRING "hello"
#set_parameter_property PWD DERIVED true
set_module_property COMPOSITION_CALLBACK elaborate

proc elaborate {} {
#	variable linesnew

#add_instance dspba my_dspba
#set_parameter_value DSPBA_MODEL_ENTITY_NAME [get_instance_parameter_value regs DSPBA_MODEL_ENTITY_NAME]

#add_instance dspba [get_instance_parameter_value regs DSPBA_MODEL_ENTITY_NAME]
#add_instance dspba [get_parameter_value DSPBA_MODEL_ENTITY_NAME]

if {[catch {add_instance dspba [get_parameter_value DSPBA_MODEL_ENTITY_NAME]} result]} {
		send_message error "Entity Name \"[get_parameter_value DSPBA_MODEL_ENTITY_NAME]\" does not exist."
} else {

#DOC_FOC_fixp16_DOC_FOC_FixP16
add_connection clk.out_clk dspba.clock
#add_connection reset.out_reset dspba.bus_reset
add_connection reset.out_reset dspba.clock_reset



#	send_message info "QV_NAME = [get_parameter_value QV_NAME]"
#	send_message info [get_instance_parameter_value regs PWD]

	foreach this_if [get_instance_interfaces dspba] {
#        if {[regexp -nocase {bus([a-z0-9_]+)} $this_if test0 test1] } {}
        if {[string compare -nocase $this_if "bus"] == 0} {
			send_message info "Adding avalon slave interface $this_if"
			add_interface $this_if avalon slave
			set_interface_property $this_if EXPORT_OF dspba.$this_if
			add_connection reset.out_reset dspba.${this_if}_reset
		}

		#set this_if_prop [get_instance_interface_property dspba $this_if interfaceType]
		set this_if_prop [get_instance_interface_properties dspba $this_if]
		#send_message info "interface $this_if = $this_if_prop"
		
	}

		#send_message info "Instances found: [get_instances]"
		#send_message info "DSPBA Interfaces found: [get_instance_interfaces "dspba"]"
		#send_message info "REG Interfaces found: [get_instance_interfaces "regs"]"
#	foreach other_if [get_instance_interfaces regs] {

		#send_message info "dspba_export_sigs = [get_instance_interface_ports regs dspba_export]"

		#if {[get_instance_parameter_value regs EXTRA_EXPORT]} {
			add_interface dspba_export conduit end
			set_interface_property dspba_export EXPORT_OF regs.dspba_export
			#send_message info "add_interface dspba_export conduit end"
		#}

		#set this_if_prop [get_instance_interface_property dspba $this_if interfaceType]
		#set this_if_prop [get_instance_interface_properties dspba $this_if]
		
#	}

	
	#add_interface slave_dspba avalon slave
	#set_interface_property slave_dspba EXPORT_OF dspba.avalon_slave_0

	add_connection regs.exp dspba.exp
}

	set_instance_parameter_value regs DSPBA_MODEL_ENTITY_NAME [get_parameter_value DSPBA_MODEL_ENTITY_NAME]
	set_instance_parameter_value regs DSPBA_MODEL_NAME [get_parameter_value DSPBA_MODEL_NAME]
	set_instance_parameter_value regs DSPBA_MODEL_PATH [get_parameter_value DSPBA_MODEL_PATH]
	set_instance_parameter_value regs QV_NAME [get_parameter_value QV_NAME]
	set_instance_parameter_value regs DV_NAME [get_parameter_value DV_NAME]
	set_instance_parameter_value regs VALID_IN_NAME [get_parameter_value VALID_IN_NAME]
	set_instance_parameter_value regs VALID_OUT_NAME [get_parameter_value VALID_OUT_NAME]
	set_instance_parameter_value regs CHANNEL_IN_NAME [get_parameter_value CHANNEL_IN_NAME]
	set_instance_parameter_value regs CHANNEL_OUT_NAME [get_parameter_value CHANNEL_OUT_NAME]
	set_instance_parameter_value regs IGNORE_NAMES [get_parameter_value IGNORE_NAMES]
	set_instance_parameter_value regs CFG_SUFFIX [get_parameter_value CFG_SUFFIX]
	set_instance_parameter_value regs FOLDING_FACTOR [get_parameter_value FOLDING_FACTOR]
	set_instance_parameter_value regs NUM_CHANNELS [get_parameter_value NUM_CHANNELS]


	#set outdir [get_generation_property OUTPUT_DIRECTORY]
	#set outputname [get_generation_property OUTPUT_NAME]  

#	send_message info "instance QV_NAME = [get_instance_parameter_value regs QV_NAME]"

}
