# This is the top level tcl file for 'FOC'

package require ::quartus::project

set overwrite_opt 0
set compile_opt 1

foreach arg $argv {
    switch $arg {
        overwrite { set overwrite_opt 1 }
        no_compile { set compile_opt 0 }
        default { puts "got unknown command: $arg" }
    }
}

if {!$overwrite_opt && [project_exists FOC]} {
    puts "The FOC quartus project already exists."
    puts "Use 'overwrite option' to force project to be over-written"
} elseif {![file exists "./rtl/DOC_FOC_fixp16_ax4/FOC/FOC.tcl"]}    {
    puts "You must run FOC.tcl "
    puts "from the ../quartus directory or else "
    puts "some of the dependant scripts won't be found."
    puts "The current directory is:" 
    puts [pwd]
    error "Exiting..."
} 

set quartus_dir $::env(QUARTUS_ROOTDIR)
puts "Creating FOC Quartus project..."

if $overwrite_opt {
    project_new -overwrite FOC
} else {
    project_new FOC
}

set_global_assignment -name FAMILY "Cyclone IV E"
set_global_assignment -name TOP_LEVEL_ENTITY DOC_FOC_fixp16_ax4_FOC
set_global_assignment -name DEVICE AUTO
set_global_assignment -name DEVICE_FILTER_SPEED_GRADE 8L
set_global_assignment -name USE_TIMEQUEST_TIMING_ANALYZER ON
set_global_assignment -name SDC_FILE ./rtl/DOC_FOC_fixp16_ax4/FOC/FOC.sdc

source ./rtl/DOC_FOC_fixp16_ax4/FOC/FOC.add.tcl

set_instance_assignment -name VIRTUAL_PIN ON -to I_Sat_Limit_cfg
set_instance_assignment -name VIRTUAL_PIN ON -to Iu
set_instance_assignment -name VIRTUAL_PIN ON -to Iw
set_instance_assignment -name VIRTUAL_PIN ON -to Ki_cfg
set_instance_assignment -name VIRTUAL_PIN ON -to Kp_cfg
set_instance_assignment -name VIRTUAL_PIN ON -to U_Sat_Limit_cfg
set_instance_assignment -name VIRTUAL_PIN ON -to Valpha
set_instance_assignment -name VIRTUAL_PIN ON -to Vbeta
set_instance_assignment -name VIRTUAL_PIN ON -to axis_in
set_instance_assignment -name VIRTUAL_PIN ON -to axis_out
set_instance_assignment -name VIRTUAL_PIN ON -to debug_1
set_instance_assignment -name VIRTUAL_PIN ON -to debug_2
set_instance_assignment -name VIRTUAL_PIN ON -to debug_3
set_instance_assignment -name VIRTUAL_PIN ON -to dv
set_instance_assignment -name VIRTUAL_PIN ON -to phi_el
set_instance_assignment -name VIRTUAL_PIN ON -to qv
set_instance_assignment -name VIRTUAL_PIN ON -to reset
set_instance_assignment -name VIRTUAL_PIN ON -to torque
set_instance_assignment -name VIRTUAL_PIN ON -to valid_in
set_instance_assignment -name VIRTUAL_PIN ON -to valid_out



if $compile_opt {
    project_archive -include_outputs -include_libraries "FOC.qar"
    package require ::quartus::flow
    puts "Running Compile flow..."
    execute_flow -compile

    package require ::quartus::report 

    # The output you get out of this script
    # is two lines of CSV formatted data (header line 1, real data line 2)
    # followed by the critical path

    load_report

    proc print_csv {expr} {
        regsub -all , $expr {} out
        regsub -all { .*} $out {} final
        if [string match $final ""] {
            # If quartus failed to find any value here, report -1
            set final -1
        }
        puts -nonewline $final
        puts -nonewline ","
    }

    create_timing_netlist -model slow
    read_sdc
    update_timing_netlist
    set clk_fmaxinfo [get_clock_fmax_info]

    puts "Extracting key data from report..."
    puts "LE,DSP_9bit,Mem_Bits,M9K,IO,FMax,Slack,Required"
    print_csv [get_fitter_resource_usage -used -le]
    print_csv [get_fitter_resource_usage -resource "Embedded Multiplier 9-bit*"]
    print_csv [get_fitter_resource_usage -used -mem_bit]
    print_csv [get_fitter_resource_usage -resource "*M9K*"]
    print_csv [get_fitter_resource_usage -used -io_pin]

    set family [get_report_panel_data -name {Flow Summary} -col 1 -row_name Family]
    set device [get_report_panel_data -name {Flow Summary} -col 1 -row_name Device]
    set fmax -1
    set restricted_fmax -1
    set clk {clk}

    foreach clkinfo $clk_fmaxinfo {
        if ([string equal $clk [lindex $clkinfo 0] ]) {
            set fmax [lindex $clkinfo 1]
            set restricted_fmax [lindex $clkinfo 2]
        }
    }

    set required_period [get_clock_info -period $clk]
    set slack [expr {$required_period - 1000.0 / $fmax}]

    print_csv $restricted_fmax
    print_csv $slack
    print_csv [expr {1000.0 / $required_period}]

    puts ""

    puts "FAMILY $family"
    puts "DEVICE $device"
    puts "TIMING PATH"
    set cpi 0
    foreach_in_collection path [get_timing_paths -npaths 1 -setup -pairs_only] {
        set from [get_node_info -name [get_path_info $path -from]]
        set to [get_node_info -name [get_path_info $path -to]]
        file delete cpath$cpi.fit
        report_timing -from_clock $clk -to_clock $clk -from $from -to $to -setup -npaths 1 -show_routing -detail full_path -file cpath$cpi.fit
        set ignore [catch {
          set f [open cpath$cpi.fit]
          puts [read $f]}]
        incr cpi
    }
}

project_close

